"""Unit tests for the run functions."""

import asyncio
import os
import sys
import unittest
import sys
from unittest.mock import call, patch, MagicMock

import sub_do

from .util import run


def mock_create_subprocess_shell(stdout="", stderr="", return_code=0):
    """Return an async method for creating the subprocess shell."""
    async def async_create_subprocess_shell(*args, **kwargs):
        """Return a mock shell."""
        class Shell:
            returncode = return_code
            async def communicate(self, *args, **kwargs):
                """Return the stdout and stderr output of the shell."""
                return stdout.encode(), stderr.encode()
        return Shell()
    return async_create_subprocess_shell


@patch.object(sys.stdout, "write")
class RunInTest(unittest.TestCase):
    """Unit tests for the run_in function."""

    @patch("asyncio.create_subprocess_shell", new=mock_create_subprocess_shell())
    def test_print_command_and_folder(self, mock_stdout_write):
        """Test that the run function prints the command and the folder on std out."""
        run(sub_do.run_in("test_folder", "test_command"))
        self.assertEqual([call("['test_command' in test_folder/ exited with 0]"), call("\n")], 
                         mock_stdout_write.call_args_list)

    @patch("asyncio.create_subprocess_shell", new=mock_create_subprocess_shell(return_code=1))
    def test_return_code(self, mock_stdout_write):
        """Test that the run function prints the (non zero) return code."""
        run(sub_do.run_in("test_folder", "test_command"))
        self.assertEqual([call("['test_command' in test_folder/ exited with 1]"), call("\n")], 
                         mock_stdout_write.call_args_list)

    @patch("asyncio.create_subprocess_shell", new=mock_create_subprocess_shell(stdout="Output on stdout"))
    def test_capture_stdout(self, mock_stdout_write):
        """Test that the run function captures and prints the stdout of the command."""
        run(sub_do.run_in("test_folder", "test_command"))
        self.assertEqual([call("['test_command' in test_folder/ exited with 0]"), call("\n"), 
                          call("[stdout]\nOutput on stdout"), call("\n")],
                         mock_stdout_write.call_args_list)

    @patch("asyncio.create_subprocess_shell", new=mock_create_subprocess_shell(stderr="Output on stderr"))
    def test_capture_stderr(self, mock_stdout_write):
        """Test that the run function captures and prints the stderr of the command."""
        run(sub_do.run_in("test_folder", "test_command"))
        self.assertEqual([call("['test_command' in test_folder/ exited with 0]"), call("\n"), 
                          call("[stderr]\nOutput on stderr"), call("\n")],
                         mock_stdout_write.call_args_list)


@patch("asyncio.create_subprocess_shell", new=mock_create_subprocess_shell())
@patch.object(sys.stdout, "write")
class RunTest(unittest.TestCase):
    """Unit tests for the run in function."""

    def test_print_command_and_folders(self, mock_stdout_write):
        """Test that the run function prints the command and the folders on std out."""
        run(sub_do.run(["folder1", "folder2"], "test_command"))
        self.assertEqual([call("['test_command' in folder1/ exited with 0]"), call("\n"),
                          call("['test_command' in folder2/ exited with 0]"), call("\n")], 
                         mock_stdout_write.call_args_list)

    def test_no_folders(self, mock_stdout_write):
        """Test that the run function prints the command and the folders on std out."""
        run(sub_do.run([], "test_command"))
        self.assertEqual([], mock_stdout_write.call_args_list)


@patch("asyncio.run", new=lambda coroutine: run(coroutine))
@patch("os.listdir", new=lambda: ["folder1", "folder2"])
@patch("os.path.isdir", new=lambda path: "folder" in path)
@patch("asyncio.create_subprocess_shell", new=mock_create_subprocess_shell())
class EntryPointTest(unittest.TestCase):
    """Unit tests for the entry point."""

    @patch("sys.argv", new=["sub-do", "test_command"])
    @patch.object(sys.stdout, "write")
    def test_print_command_and_folders(self, mock_stdout_write):
        """Test that the run function prints the command and the folders on standard out."""
        sub_do.sub_do()
        self.assertEqual([call("['test_command' in folder1/ exited with 0]"), call("\n"),
                          call("['test_command' in folder2/ exited with 0]"), call("\n")], 
                         mock_stdout_write.call_args_list)

    @patch("sys.argv", new=["sub-do"])
    @patch.object(sys.stderr, "write")
    def test_no_arguments(self, mock_stderr_write):
        """Test that the run function prints the command and the folders on standard out."""
        self.assertRaises(SystemExit, sub_do.sub_do)
        self.assertEqual([call("usage: sub-do [-h] [-V] [-d [directory [directory ...]]] command\n"),
                          call("sub-do: error: the following arguments are required: command\n")], 
                         mock_stderr_write.call_args_list)
